;;;
;;; Copyright (c) 1989, 1993, 1994
;;;	The Regents of the University of California.  All rights reserved.
;;;
;;; This code is derived from software contributed to Berkeley by
;;; Michael Fischbein.
;;;
;;; Redistribution and use in source and binary forms, with or without
;;; modification, are permitted provided that the following conditions
;;; are met:
;;; 1. Redistributions of source code must retain the above copyright
;;;    notice, this list of conditions and the following disclaimer.
;;; 2. Redistributions in binary form must reproduce the above copyright
;;;    notice, this list of conditions and the following disclaimer in the
;;;    documentation and/or other materials provided with the distribution.
;;; 4. Neither the name of the University nor the names of its contributors
;;;    may be used to endorse or promote products derived from this software
;;;    without specific prior written permission.
;;;
;;; THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
;;; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
;;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
;;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
;;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
;;; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
;;; HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
;;; LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
;;; OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
;;; SUCH DAMAGE.
;;;

(use gauche.experimental.lamb)

(.include <sys/cdefs.h> <sys/types.h> <sys/stat.h> <sys/ioctl.h>
          <proc/pwcache.h> <dirent.h> <err.h> <errno.h>
          <fts.h> <grp.h> <inttypes.h> <limits.h> <locale.h>
          <pwd.h> <stdio.h> <stdlib.h> <string.h> <unistd.h>)
(.include "ls.h" "extern.h")

;;
;; Upward approximation of the maximum number of characters needed to
;; represent a value of integral type t as a string, excluding the
;; NUL terminator, with provision for a sign.
;;
(define-cise-expr (STRBUF_SIZEOF t) `(+ 1 (/ (* CHAR_BIT (sizeof ,t)) 3) 1))

;;
;; MAKENINES(n) turns n into (10**n)-1.  This is useful for converting a width
;; into a number that wide in decimal.
;; XXX: Overflows are not considered.
;;
(define-cise-expr (MAKENINES n)
  `(let* ([i::intmax_t 1])
     (for [() (> ,n 0) (*= i 10)] (dec! ,n))
     (= ,n (- i 1))))

;; Options DSL
(define *flags* '())
(define *options* (make-hash-table 'eqv?))
(define-cise-expr (define-flag type name on off)
  (push! *flags* #`"extern ,type ,name;\n")
  (dolist [c (string->list on)]  (hash-table-push! *options* c (cons name 1)))
  (dolist [c (string->list off)] (hash-table-push! *options* c (cons name 0)))
  `(,#`",type ,name"))
(define-cise-stmt (emit-option-switch ch)
  `(case ,ch
     ,@(hash-table-map *options*
        (lambda (k v) `((,k) ,@(map (lambda (x) `(=,(car x),(cdr x))) v))))))
   
(.raw-c-code
"static void (*printfcn)(const DISPLAY *);
static int (*sortfcn)(const FTSENT *, const FTSENT *);

long blocksize;			/* block size units */
int termwidth = 80;		/* default terminal width */

static int rval;
static char dot[] = \".\", *dotav[] = {dot, NULL};")

(define-flag int f_accesstime "u" "cU") ; use time of last access
(define-flag int f_inode      "i" "")   ; print inode
(define-flag int f_kblocks    "k" "")   ; print size in kilobytes
(define-flag int f_listdir    "d" "")   ; list actual directory, not contents
(define-flag int f_listdot    "aA" "")  ; list files beginning with .
(define-flag int f_noautodot  "I" "")   ; do not automatically enable -A for root
(define-flag int f_longform   "l" "1Cxm"); long listing format
(define-flag int f_nonprint   "q" "Bbw"); show unprintables as ?
(define-flag int f_nosort     "f" "")   ; don't sort output
(define-flag int f_notabs     "" "")    ; don't use tab-separated multi-col output
(define-flag int f_numericonly "n" "")  ; don't convert uid/gid to name
(define-flag int f_octal      "B" "qbw"); show unprintables as \\xxx
(define-flag int f_octal_escape "b" "Bqw"); like f_octal but use C escapes if possible
(define-flag int f_recursive  "R" "d")  ; ls subdirectories also
(define-flag int f_reversesort "r" "")  ; reverse whatever sort is used
(define-flag int f_sectime    "T" "")   ; print the real time for all files
(define-flag int f_singlecol  "1" "Clxm"); use single column output
(define-flag int f_size       "s" "")   ; list size in short listing
(define-flag int f_slash      "p" "F")  ; similar to f_type, but only for dirs
(define-flag int f_sortacross "x" "C")  ; sort across rows, not down columns
(define-flag int f_statustime "c" "uU") ; use time of last mode change
(define-flag int f_stream     "m" "1l") ; stream the output, separate with commas
(define-flag int f_timesort   "t" "S")  ; sort by time vice name
(define-flag int f_sizesort   "S" "t")
(define-flag int f_type       "Fp" "")  ; add type character for non-regular files
(define-flag int f_whiteout   "W" "")   ; show whiteout entries

(.static-decls)

(define-cfn main (argc::int argv::char**) ::int
  (setlocale LC_ALL "")
  ;; Terminal defaults to -Cq, non-terminal defaults to -1. 
  (cond [(isatty STDOUT_FILENO)
         (= termwidth 80)
         (let* ([p::char* (getenv "COLUMNS")]
                [win::(struct winsize)])
           (cond [(and p (!= (* p) #\null)) (= termwidth (atoi p))]
                 [(and (!= (ioctl STDOUT_FILENO TIOCGWINSZ (& win)) -1)
                       (> (ref win ws_col) 0))
                  (= termwidth (ref win ws_col))])
           (= f_nonprint 1))]
        [else
         (= f_singlecol 1)
         ;; retrieve environment variable, in case of explicit -C
         (let* ([p::char* (getenv "COLUMMS")])
           (when p (= termwidth (atoi p))))])

  (let* ([ch::int] [fts_options::int FTS_PHYSICAL])
    (while (!= (= ch (getopt argc argv "1ABCFGHILPRSTUWZabcdfghiklmnopqrstuwx")) -1)
      (emit-option-switch ch)
      (case ch
        [(#\H) (logior= fts_options FTS_COMFOLLOW)]
        [(#\L) (logand= fts_options (lognot FTS_PHYSICAL))
               (logior= fts_options FTS_LOGICAL)]
        [(#\P) (logand= fts_options (lognot FTS_COMFOLLOW))
               (logand= fts_options (lognot FTS_LOGICAL))
               (logior= fts_options FTS_PHYSICAL)]
        [(#\a) (logior= fts_options FTS_SEEDOT)]
        [(#\g)]                         ; Compatibility with 4.3BSD.
        [(#\?) (usage)]))
    (-= argc optind)
    (+= argv optind)

    ;; Root is -A automatically unless -I.
    (when (and (not f_listdot) (== (getuid) (cast uid_t 0)) (not f_noautodot))
      (= f_listdot 1))

    (when (and (getenv "CLICOLOR")
               (or (isatty STDOUT_FILENO) (getenv "CLICOLOR_FORCE")))
      (warnx "color support not compiled in"))

    ;; If not -F, -i, -l, -s, -S or -t options, don't require stat
    ;; information, unless in color mode in which case we do
    ;; need this to determine which colors to display.
    (when (and (not f_inode) (not f_longform) (not f_size) (not f_timesort)
               (not f_sizesort) (not f_type))
      (logior= fts_options FTS_NOSTAT))
    
    ;; If not -F, -d or -l options, follow any symbolic links listed on
    ;; the command line.
    (when (and (not f_longform) (not f_listdir) (not f_type))
      (logior= fts_options FTS_COMFOLLOW))
        
    ;; If -l or -s, figure out block size.
    (when (or f_longform f_size)
      (if f_kblocks (= blocksize 2) (= blocksize 1)))

    ;; Select a sort function
    (if f_reversesort
      (cond [(and (not f_timesort) (not f_sizesort)) (= sortfcn revnamecmp)]
            [f_sizesort                              (= sortfcn revsizecmp)]
            [f_accesstime                            (= sortfcn revacccmp)]
            [f_statustime                            (= sortfcn revstatcmp)]
            [else                                    (= sortfcn revmodcmp)])
      (cond [(and (not f_timesort) (not f_sizesort)) (= sortfcn namecmp)]
            [f_sizesort                              (= sortfcn sizecmp)]
            [f_accesstime                            (= sortfcn acccmp)]
            [f_statustime                            (= sortfcn statcmp)]
            [else                                    (= sortfcn modcmp)]))

    ;; Select a print function
    (cond [f_singlecol  (= printfcn printscol)]
          [f_longform   (= printfcn printlong)]
          [f_stream     (= printfcn printstream)]
          [else         (= printfcn printcol)])

    (if argc (traverse argc argv fts_options) (traverse 1 dotav fts_options))
    (exit rval)))

(.raw-c-code "static int output; /* If anything output. */")

;; Traverse() walks the logical directory structure specified by the argv list
;; in the order specified by the mastercmp() comparison function.  During the
;; traversal it passes linked lists of structures to display() which represent
;; a superset (may be exact set) of the files to be displayed.

(define-cfn traverse (argc::int argv::char** options::int) ::void :static
  (let* ([ftsp::FTS* (fts_open argv options (?: f_nosort NULL mastercmp))])
    (when (== ftsp NULL) (err 1 "fts_open"))
    ;; We ignore errors from fts_children here since they will be
    ;; replicated and signalled on the next call to fts_read() below.
    (let* ([chp::FTSENT* (fts_children ftsp 0)])
      (when (!= chp NULL) (display NULL chp options))
      (when f_listdir (return))
      ;; If not recursing down this tree and don't need stat info, just get
      ;; the names.
      (let* ([ch_options::int (and (not f_recursive)
                                   (logand options (?: FTS_NOSTAT FTS_NAMEONLY 0)))]
             [p::FTSENT*])
        (while (!= (= p (fts_read ftsp)) NULL)
          (case (-> p fts_info)
            [(FTS_DC) (warnx "%s: directory causes a cycle" (-> p fts_name))]
            [(FTS_DNR FTS_ERR)
             (warnx "%s: %s" (-> p fts_name) (strerror (-> p fts_errno)))
             (= rval 1)]
            [(FTS_D)
             (when (and (!= (-> p fts_level) FTS_ROOTLEVEL)
                        (== (aref (-> p fts_name) 0) #\.)
                        (not f_listdot))
               (break))
             ;; If already output something, put out a newline as
             ;; a separator.  If multiple arguments, precede each
             ;; directory with its name.
             (cond [output (putchar #\newline)
                           (printname (-> p fts_path))
                           (puts ":")]
                   [(> argc 1) (printname (-> p fts_path))
                               (puts ":")
                               (= output 1)])
             (= chp (fts_children ftsp ch_options))
             (display p chp options)
             (unless (and f_recursive (!= chp NULL))
               (fts_set ftsp p FTS_SKIP))
             (break)]
            [else (break)]))
        (when errno (err 1 "fts_read"))))))

(define-cise-stmt set-if-gt!
  [(_ var ('= tmp expr)) `(when (> (= ,tmp ,expr) ,var) (= ,var ,tmp))]
  [(_ var val) `(when (> ,val ,var) (= ,var ,val))])

(define-cise-stmt (fill-struct obj . spec)
  `(begin ,@(map (^.[(s v)`(= (-> ,obj ,s) ,v)][n`(= (-> ,obj ,n) ,n)]) spec)))

;;
;; Display() takes a linked list of FTSENT structures and passes the list
;; along with any other necessary information to the print function.  P
;; points to the parent directory of the display list.
;;

(define-cfn display (p::(const FTSENT*) list::FTSENT* options::int)
  ::void :static
  (let* ([initmax::char* (getenv "LS_COLWIDTHS")]
         [maxblock::long 0] [maxinode::u_long 0] [maxlen::u_long 0]
         [maxnlink::u_long 0] [maxuser::uid_t 0] [maxgroup::gid_t 0]
         [maxsize::off_t 0])
    (when (and (!= initmax NULL) (!= (* initmax) #\null))
      (let* ([jinitmax::char* (malloc (+ (* (strlen initmax) 2) 2))])
        (when (== jinitmax NULL) (err 1 "malloc"))
        (let* ([initmax2::char* jinitmax])
          (cond [(== (* initmax) #\:) (strcpy initmax2 "0:") (+= initmax2 2)]
                [else                 (set! (* (post++ initmax2)) (* initmax)
                                            (* initmax2) #\null)])
          (for [(inc! initmax) (!= (* initmax) #\null) (inc! initmax)]
               (cond [(and (== (aref initmax -1) #\:)
                           (== (aref initmax 0) #\:))
                      (set! (* (post++ initmax2)) #\0
                            (* (post++ initmax2)) (aref initmax 0)
                            (aref initmax2 1) #\null)]
                     [else
                      (set! (* (post++ initmax2)) (aref initmax 0)
                            (aref initmax2 1) #\null)]))
          (when (== (aref initmax2 -1) #\:) (strcpy initmax2 "0"))
          (let* ([ninitmax::int
                  (sscanf jinitmax " %lu : %ld : %lu : %u : %u : %lu : %lu "
                          (& maxinode) (& maxblock) (& maxnlink)
                          (& maxuser) (& maxgroup) (& maxsize) (& maxlen))])
            (= f_notabs 1)
            (case/fallthrough ninitmax
                              [(0) (= maxinode 0)]
                              [(1) (= maxblock 0)]
                              [(2) (= maxnlink 0)]
                              [(3) (= maxuser 0)]
                              [(4) (= maxgroup 0)]
                              [(5) (= maxsize 0)]
                              [(6) (= maxlen 0) (= f_notabs 0)])
            (MAKENINES maxinode)
            (MAKENINES maxblock)
            (MAKENINES maxnlink)
            (MAKENINES maxsize)
            (free jinitmax)))))
    (let* ([cur::FTSENT*] [entries::int 0] [bcfile::int 0] [btotal::int 0]
           [needstats::int (or f_inode f_longform f_size)])
      (for [(= cur list) cur (= cur (-> cur fts_link))]
           (when (or (== (-> cur fts_info) FTS_ERR)
                     (== (-> cur fts_info) FTS_NS))
             (warnx "%s: %s" (-> cur fts_name) (strerror (-> cur fts_errno)))
             (= (-> cur fts_number) NO_PRINT)
             (= rval 1)
             (continue))
           ;; P is NULL if list is the argv list, to which different rules
           ;; apply.
           (if (== p NULL)
             (when (and (== (-> cur fts_info) FTS_D) (not f_listdir))
               (= (-> cur fts_number) NO_PRINT) (continue))
             ;; Only display dot file if -a/-A set.
             (when (and (== (aref (-> cur fts_name) 0) #\.) (not f_listdot))
               (= (-> cur fts_number) NO_PRINT) (continue)))
           (set-if-gt! maxlen (-> cur fts_namelen))
           (when (or f_octal f_octal_escape)
             (let* ([t::u_long (len_octal (-> cur fts_name) (-> cur fts_namelen))])
               (set-if-gt! maxlen t)))
           (when needstats
             (let* ([sp::(struct stat*) (-> cur fts_statp)]
                    [user::(const char*)] [group::(const char*)]
                    [ulen::size_t] [glen::size_t]
                    [nuser::(.array char ((+ (STRBUF_SIZEOF uid_t) 1)))]
                    [ngroup::(.array char ((+ (STRBUF_SIZEOF gid_t) 1)))])
               (set-if-gt! maxblock (-> sp st_blocks))
               (set-if-gt! maxinode (-> sp st_ino))
               (set-if-gt! maxnlink (-> sp st_nlink))
               (set-if-gt! maxsize  (-> sp st_size))
               (+= btotal (-> sp st_blocks))
               (when f_longform
                 (cond [f_numericonly
                        (snprintf nuser (sizeof nuser) "%u" (-> sp st_uid))
                        (snprintf ngroup (sizeof ngroup) "%u" (-> sp st_gid))
                        (set! user nuser group ngroup)]
                       [else
                        (set! user  (user_from_uid (-> sp st_uid))
                              group (group_from_gid (-> sp st_gid)))])
                 (set-if-gt! maxuser  (= ulen (strlen user)))
                 (set-if-gt! maxgroup (= glen (strlen group)))
                 (let* ([np::NAMES* (malloc (+ (sizeof NAMES) ulen glen 4))])
                   (when (== np NULL) (err 1 "malloc"))
                   (= (-> np user) (& (aref (-> np data) 0)))
                   (strcpy (-> np user) user)
                   (= (-> np group) (& (aref (-> np data) (+ ulen 1))))
                   (strcpy (-> np group) group)
                   (when (or (S_ISCHR (-> sp st_mode)) (S_ISBLK (-> sp st_mode)))
                     (= bcfile 1))
                   (= (-> cur fts_pointer) np)))))
           (pre++ entries))
      ;; If there are no entries to display, we normally stop right
      ;; here.  However, we must continue if we have to display the
      ;; total block count.  In this case, we display the total only
      ;; on the second (p != NULL) pass.
      (when (and (not entries) (or (not (or f_longform f_size)) (== p NULL)))
        (return))

      (let* ([d::DISPLAY] [buf::(.array char ((+ (STRBUF_SIZEOF u_quad_t) 1)))])
        (fill-struct (& d) list entries maxlen)
        (when needstats
          (fill-struct (& d) bcfile btotal (s_group maxgroup) (s_user maxuser))
          (snprintf buf (sizeof buf) "%lu" maxblock)
          (set! (ref d s_block) (strlen buf))
          (snprintf buf (sizeof buf) "%lu" maxinode)
          (set! (ref d s_inode) (strlen buf))
          (snprintf buf (sizeof buf) "%lu" maxnlink)
          (set! (ref d s_nlink) (strlen buf))
          (snprintf buf (sizeof buf) "%lu" maxsize)
          (set! (ref d s_size) (strlen buf)))
        (printfcn (& d))
        (= output 1)
        (when f_longform
          (for [(= cur list) cur (= cur (-> cur fts_link))]
               (free (-> cur fts_pointer))))))))
             
;;
;; Ordering for mastercmp:
;; If ordering the argv (fts_level = FTS_ROOTLEVEL) return non-directories
;; as larger than directories.  Within either group, use the sort function.
;; All other levels use the sort function.  Error entries remain unsorted.
;;
(define-cfn mastercmp (a::(const FTSENT **) b::(const FTSENT **)) ::int :static
  (let* ([a_info::int (-> (* a) fts_info)]
         [b_info::int (-> (* b) fts_info)])
    (when (== a_info FTS_ERR) (return 0))
    (when (== b_info FTS_ERR) (return 0))
    (when (or (== a_info FTS_NS) (== b_info FTS_NS))
      (return (namecmp (* a) (* b))))
    (when (and (!= a_info b_info)
               (== (-> (* a) fts_level) FTS_ROOTLEVEL)
               (not f_listdir))
      (when (== a_info FTS_D) (return 1))
      (when (== b_info FTS_D) (return -1)))
    (return (sortfcn (* a) (* b)))))

;; emit flags.h
(with-output-to-file "flags.h" (cut for-each (cut display <>) *flags*))

;; Local variables:
;; mode: scheme
;; end:
