;;;
;;; Copyright (c) 1989, 1993, 1994
;;;	The Regents of the University of California.  All rights reserved.
;;;
;;; This code is derived from software contributed to Berkeley by
;;; Michael Fischbein.
;;;
;;; Redistribution and use in source and binary forms, with or without
;;; modification, are permitted provided that the following conditions
;;; are met:
;;; 1. Redistributions of source code must retain the above copyright
;;;    notice, this list of conditions and the following disclaimer.
;;; 2. Redistributions in binary form must reproduce the above copyright
;;;    notice, this list of conditions and the following disclaimer in the
;;;    documentation and/or other materials provided with the distribution.
;;; 4. Neither the name of the University nor the names of its contributors
;;;    may be used to endorse or promote products derived from this software
;;;    without specific prior written permission.
;;;
;;; THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
;;; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
;;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
;;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
;;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
;;; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
;;; HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
;;; LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
;;; OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
;;; SUCH DAMAGE.
;;;

(.include  <sys/cdefs.h> <sys/param.h> <sys/stat.h>
           <err.h> <errno.h> <fts.h> <langinfo.h>
           <stdio.h> <stdlib.h> <string.h> <time.h> <unistd.h>)
(.include "ls.h" "extern.h")
(.static-decls)

(define-cise-stmt (do-ftsent (ent start) . body)
  `(let* ([,ent :: FTSENT* ,start])
     (for [() ,ent (= ,ent (-> ,ent fts_link))] ,@body)))

(define-cise-expr (IS_NOPRINT p) `(== (-> ,p fts_number) NO_PRINT))

(define-cfn printscol (dp::(const DISPLAY *)) ::void
  (do-ftsent [p (-> dp list)]
    (when (IS_NOPRINT p) (continue))
    (printaname p (-> dp s_inode) (-> dp s_block))
    (putchar #\newline)))

;; print name in current style
(define-cfn printname (name::(const char*)) ::int
  (cond [(or f_octal f_octal_escape) (return (prn_octal name))]
        [f_nonprint                  (return (prn_printable name))]
        [else                        (return (prn_normal name))]))

(define-cfn printlong (dp::(const DISPLAY*)) ::void
  (when (and (or (== (-> dp list) NULL)
                 (!= (-> dp list fts_level) FTS_ROOTLEVEL))
             (or f_longform f_size))
    (printf "total %lu\n" (howmany (-> dp btotal) blocksize)))
  (do-ftsent [p (-> dp list)]
    (when (IS_NOPRINT p) (continue))
    (let* ([sp::(struct stat*) (-> p fts_statp)]
           [buf::(.array char (20))])
      (when f_inode
        (printf "%*lu " (-> dp s_inode) (cast u_long (-> sp st_ino))))
      (when f_size
        (printf "%*ld " (-> dp s_block) (howmany (-> sp st_blocks) blocksize)))
      (strmode (-> sp st_mode) buf)
      (let* ([np::NAMES* (-> p fts_pointer)])
        (printf "%s %*u %-*s  %-*s  " buf (-> dp s_nlink) (-> sp st_nlink)
                (-> dp s_user) (-> np user) (-> dp s_group) (-> np group))
        (cond [(or (S_ISCHR (-> sp st_mode)) (S_ISBLK (-> sp st_mode)))
               (if (or (> (minor (-> sp st_rdev)) 255)
                       (< (minor (-> sp st_rdev)) 0))
                 (printf "%3d, 0x%08x " (major (-> sp st_rdev))
                         (cast u_int (minor (-> sp st_rdev))))
                 (printf "%3d, %3d " (major (-> sp st_rdev)) (minor (-> sp st_rdev))))]
              [(-> dp bcfile) (printf "%*s%*ld " (- 8 (-> dp s_size)) ""
                                      (-> dp s_size) (-> sp st_size))]
              [else (printsize (-> dp s_size) (-> sp st_size))])
        (cond [f_accesstime (printtime (-> sp st_atime))]
              [f_statustime (printtime (-> sp st_ctime))]
              [else         (printtime (-> sp st_mtime))])
        (printname (-> p fts_name))
        (when f_type (printtype (-> sp st_mode)))
        (when (S_ISLNK (-> sp st_mode)) (printlink p))
        (putchar #\newline)))
    ))

(define-cfn printstream (dp::(const DISPLAY*)) ::void
  (let* ([chcnt::int 0])
    (do-ftsent [p (-> dp list)]
      (when (== (-> p fts_number) NO_PRINT) (continue))
      ;;XXX strlen does not take octal escapes into account.
      (when (>= (+ (strlen (-> p fts_name)) chcnt (?: (-> p fts_link) 2 0))
                (cast unsigned termwidth))
        (putchar #\newline) (= chcnt 0))
      (+= chcnt (printaname p (-> dp s_inode) (-> dp s_block)))
      (when (-> p fts_link) (printf ", ") (+= chcnt 2)))
    (when chcnt (putchar #\newline))))

(define-cfn printcol (dp::(const DISPLAY*)) ::void
  (let* ([tabwidth::int (?: f_notabs 1 8)]
         [lastentries::(static int) -1]
         [array::(static FTSENT**)]
         [narray::FTSENT**]
         [num::int 0])
    ;; Have to do random access in the linked list -- build a table
    ;; of pointers.
    (when (> (-> dp entries) lastentries)
      (when (== (= narray (realloc array (* (-> dp entries) (sizeof (FTSENT*)))))
                NULL)
        (warn NULL) (printscol dp) (return))
      (= lastentries (-> dp entries))
      (= array narray))
    (do-ftsent [p (-> dp list)]
      (when (!= (-> p fts_number) NO_PRINT) (= (aref array (post++ num)) p)))

    (let* ([colwidth::int (-> dp maxlen)])
      (when f_inode (+= colwidth (+ (-> dp s_inode) 1)))
      (when f_size  (+= colwidth (+ (-> dp s_block) 1)))
      (when f_type  (+= colwidth 1))
      (= colwidth (logand (+ colwidth tabwidth) (lognot (- tabwidth 1))))
      (when (< termwidth (* 2 colwidth)) (printscol dp) (return))

      (let* ([numcols::int (/ termwidth colwidth)]
             [numrows::int (/ num numcols)])
        (when (% num numcols) (inc! numrows))
        (when (and (or (== (-> dp list) NULL)
                       (!= (-> dp list fts_level) FTS_ROOTLEVEL))
                   (or f_longform f_size))
          (printf "total %lu\n" (howmany (-> dp btotal) blocksize)))

        (let* ([base::int 0])
          (dotimes [row numrows]
            (let* ([endcol::int colwidth] [chcnt::int 0] [cnt::int])
              (when (not f_sortacross) (= base row))
              (dotimes [col numcols]
                (+= chcnt (printaname (aref array base) (-> dp s_inode)
                                      (-> dp s_block)))
                (if f_sortacross (inc! base) (+= base numrows))
                (when (>= base num) (break))
                (while (<= (= cnt (logand (+ chcnt tabwidth) (lognot (- tabwidth 1))))
                           endcol)
                  (when (and f_sortacross (>= (+ col 1) numcols)) (break))
                  (putchar (?: f_notabs #\space #\tab))
                  (= chcnt cnt))
                (+= endcol colwidth))
              (putchar #\newline))))
        ))))
        
;; print [inode] [size] name
;; return # of characters printed, no trailing characters.
(define-cfn printaname (p::(const FTSENT*) inodefield::u_long sizefield::u_long)
  ::int :static
  (let* ([sp::(struct stat*) (-> p fts_statp)] [chcnt::int 0])
    (when f_inode
      (+= chcnt (printf "%*lu " (cast int inodefield) (cast u_long (-> sp st_ino)))))
    (when f_size
      (+= chcnt (printf "%*ld "	(cast int sizefield)
                        (howmany (-> sp st_blocks) blocksize))))
    (+= chcnt (printname (-> p fts_name)))
    (when f_type (+= chcnt (printtype (-> sp st_mode))))
    (return chcnt)))

(define-cise-expr (SIXMONTHS) (* (/ 365 2) 86400))

(define-cfn printtime (ftime::time_t) ::void :static
  (let* ([d_first::(static int) 1] [now::(static time_t) 0]
         [format::(const char*)])
    (when (== now 0) (= now (time NULL)))
    (cond [f_sectime (= format (?: d_first "%e %b %T %Y " "%b %e %T %Y "))]
          [(and (> (+ ftime (SIXMONTHS)) now) (< ftime (+ now (SIXMONTHS))))
           (= format (?: d_first "%e %b %R " "%b %e %R "))]
          [else (= format (?: d_first "%e %b  %Y " "%b %e  %Y "))])
    (let* ([longstring::(.array char (80))])
      (strftime longstring (sizeof longstring) format (localtime (& ftime)))
      (fputs longstring stdout))))

(define-cfn printtype (mode::u_int) ::int :static
  (when f_slash
    (when (== (logand mode S_IFMT) S_IFDIR) (putchar #\/) (return 1))
    (return 0))
  (case (logand mode S_IFMT)
    [(S_IFDIR) (putchar #\/) (return 1)]
    [(S_IFIFO) (putchar #\|) (return 1)]
    [(S_IFLNK) (putchar #\@) (return 1)]
    [(S_IFSOCK) (putchar #\=) (return 1)]
    [else (break)])
  (when (logand mode (logior S_IXUSR (logior S_IXGRP S_IXOTH)))
    (putchar #\*) (return 1))
  (return 0))

(define-cfn printlink (p::(const FTSENT*)) ::void :static
  (let* ([name::(.array char ((+ MAXPATHLEN 1)))]
         [path::(.array char ((+ MAXPATHLEN 1)))])
    (if (== (-> p fts_level) FTS_ROOTLEVEL)
      (snprintf name (sizeof name) "%s" (-> p fts_name))
      (snprintf name (sizeof name) "%s/%s"
                (-> p fts_parent fts_accpath) (-> p fts_name)))
    (let* ([lnklen::int (readlink name path (- (sizeof path) 1))])
      (when (== lnklen -1)
        (fprintf stderr "\nls: %s: %s\n" name (strerror errno))
        (return))
      (= (aref path lnklen) #\null)
      (printf " -> ")
      (printname path))))

(define-cfn printsize (width::size_t bytes::off_t) ::void :static
  (printf "%*ld " (cast u_int width) bytes))

(define-cfn strmode (mode::mode_t buf::char*) ::void :static
  (case (logand mode S_IFMT)
    [(S_IFDIR)  (= (* buf) #\d)]
    [(S_IFLNK)  (= (* buf) #\l)]
    [(S_IFSOCK) (= (* buf) #\s)]
    [(S_IFIFO)  (= (* buf) #\p)]
    [(S_IFBLK)  (= (* buf) #\b)]
    [(S_IFCHR)  (= (* buf) #\c)]
    [else       (= (* buf) #\-)])
  (= (aref buf 1) (?: (logand mode S_IRUSR) #\r #\-)
     (aref buf 2) (?: (logand mode S_IWUSR) #\w #\-)
     (aref buf 3) (?: (logand mode S_IXUSR)
                      (?: (logand mode S_ISUID) #\s #\x)
                      (?: (logand mode S_ISUID) #\S #\-))
     (aref buf 4) (?: (logand mode S_IRGRP) #\r #\-)
     (aref buf 5) (?: (logand mode S_IWGRP) #\w #\-)
     (aref buf 6) (?: (logand mode S_IXGRP)
                      (?: (logand mode S_ISGID) #\s #\x)
                      (?: (logand mode S_ISGID) #\S #\-))
     (aref buf 7) (?: (logand mode S_IROTH) #\r #\-)
     (aref buf 8) (?: (logand mode S_IWOTH) #\w #\-)
     (aref buf 9) (?: (logand mode S_IXOTH)
                      (?: (logand mode S_ISVTX) #\t #\x)
                      (?: (logand mode S_ISVTX) #\T #\-))
     (aref buf 10) #\null))
